@@ -2,7 +2,7 @@ import html
 import json
 import string
 import warnings
-from urllib.parse import quote, urlparse
+from urllib.parse import urlparse
 
 from django.conf import settings
 from django.contrib import messages
@@ -32,17 +32,11 @@ from django.utils.encoding import force_str
 from django.utils.translation import gettext_lazy as _
 
 from allauth import app_settings as allauth_app_settings
-from allauth.account import signals
-from allauth.account.app_settings import AuthenticationMethod
+from allauth.account import app_settings, signals
 from allauth.core import context, ratelimit
 from allauth.core.internal.adapter import BaseAdapter
-from allauth.utils import (
-    build_absolute_uri,
-    generate_unique_username,
-    import_attribute,
-)
-
-from . import app_settings
+from allauth.core.internal.httpkit import headed_redirect_response
+from allauth.utils import generate_unique_username, import_attribute
 
 
 class DefaultAccountAdapter(BaseAdapter):
@@ -55,6 +49,9 @@ class DefaultAccountAdapter(BaseAdapter):
 
     error_messages = {
         "account_inactive": _("This account is currently inactive."),
+        "cannot_remove_primary_email": _(
+            "You cannot remove your primary email address."
+        ),
         "duplicate_email": _(
             "This email address is already associated with this account."
         ),
@@ -65,12 +62,14 @@ class DefaultAccountAdapter(BaseAdapter):
         "enter_current_password": _("Please type your current password."),
         "incorrect_code": _("Incorrect code."),
         "incorrect_password": _("Incorrect password."),
+        "invalid_or_expired_key": _("Invalid or expired key."),
         "invalid_password_reset": _("The password reset token was invalid."),
         "max_email_addresses": _("You cannot add more than %d email addresses."),
         "too_many_login_attempts": _(
             "Too many failed login attempts. Try again later."
         ),
         "unknown_email": _("The email address is not assigned to any user account"),
+        "unverified_primary_email": _("Your primary email address must be verified."),
         "username_blacklisted": _(
             "Username can not be used. Please use other username."
         ),
@@ -90,12 +89,6 @@ class DefaultAccountAdapter(BaseAdapter):
         request.session["account_verified_email"] = None
         return ret
 
-    def stash_user(self, request, user):
-        request.session["account_user"] = user
-
-    def unstash_user(self, request):
-        return request.session.pop("account_user", None)
-
     def is_email_verified(self, request, email):
         """
         Checks whether or not the email address is already verified
@@ -108,7 +101,10 @@ class DefaultAccountAdapter(BaseAdapter):
             ret = verified_email.lower() == email.lower()
         return ret
 
-    def can_delete_email(self, email_address):
+    def can_delete_email(self, email_address) -> bool:
+        """
+        Returns whether or not the given email address can be deleted.
+        """
         from allauth.account.models import EmailAddress
 
         has_other = (
@@ -117,7 +113,8 @@ class DefaultAccountAdapter(BaseAdapter):
             .exists()
         )
         login_by_email = (
-            app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.EMAIL
+            app_settings.AUTHENTICATION_METHOD
+            == app_settings.AuthenticationMethod.EMAIL
         )
         if email_address.primary:
             if has_other:
@@ -137,7 +134,10 @@ class DefaultAccountAdapter(BaseAdapter):
         else:
             return True
 
-    def format_email_subject(self, subject):
+    def format_email_subject(self, subject) -> str:
+        """
+        Formats the given email subject.
+        """
         prefix = app_settings.EMAIL_SUBJECT_PREFIX
         if prefix is None:
             site = get_current_site(context.request)
@@ -209,7 +209,7 @@ class DefaultAccountAdapter(BaseAdapter):
         that URLs passed explicitly (e.g. by passing along a `next`
         GET parameter) take precedence over the value returned here.
         """
-        assert request.user.is_authenticated
+        assert request.user.is_authenticated  # nosec
         url = getattr(settings, "LOGIN_REDIRECT_URLNAME", None)
         if url:
             warnings.warn(
@@ -230,15 +230,20 @@ class DefaultAccountAdapter(BaseAdapter):
         """
         return resolve_url(app_settings.LOGOUT_REDIRECT_URL)
 
-    def get_email_confirmation_redirect_url(self, request):
+    def get_email_verification_redirect_url(self, email_address):
         """
-        The URL to return to after successful email confirmation.
+        The URL to return to after email verification.
         """
-        if request.user.is_authenticated:
+        get_url = getattr(self, "get_email_confirmation_redirect_url", None)
+        if get_url:
+            # Deprecated.
+            return get_url(self.request)
+
+        if self.request.user.is_authenticated:
             if app_settings.EMAIL_CONFIRMATION_AUTHENTICATED_REDIRECT_URL:
                 return app_settings.EMAIL_CONFIRMATION_AUTHENTICATED_REDIRECT_URL
             else:
-                return self.get_login_redirect_url(request)
+                return self.get_login_redirect_url(self.request)
         else:
             return app_settings.EMAIL_CONFIRMATION_ANONYMOUS_REDIRECT_URL
 
@@ -378,6 +383,8 @@ class DefaultAccountAdapter(BaseAdapter):
         Wrapper of `django.contrib.messages.add_message`, that reads
         the message text from a template.
         """
+        if getattr(getattr(request, "allauth", None), "headless", None):
+            return
         if "django.contrib.messages" in settings.INSTALLED_APPS:
             if message:
                 messages.add_message(request, level, message, extra_tags=extra_tags)
@@ -518,11 +525,14 @@ class DefaultAccountAdapter(BaseAdapter):
         """
         Marks the email address as confirmed on the db
         """
-        from allauth.account.internal.flows import manage_email
+        from allauth.account.internal.flows import email_verification
 
-        return manage_email.confirm_email(request, email_address)
+        return email_verification.verify_email(request, email_address)
 
-    def set_password(self, user, password):
+    def set_password(self, user, password) -> None:
+        """
+        Sets the password for the user.
+        """
         user.set_password(password)
         user.save()
 
@@ -556,19 +566,24 @@ class DefaultAccountAdapter(BaseAdapter):
 
         return url_has_allowed_host_and_scheme(url, allowed_hosts=allowed_hosts)
 
+    def send_password_reset_mail(self, user, email, context):
+        """
+        Method intended to be overridden in case you need to customize the logic
+        used to determine whether a user is permitted to request a password reset.
+        For example, if you are enforcing something like "social only" authentication
+        in your app, you may want to intervene here by checking `user.has_usable_password`
+
+        """
+        return self.send_mail("account/email/password_reset_key", email, context)
+
     def get_reset_password_from_key_url(self, key):
         """
-        Method intented to be overriden in case the password reset email
-        needs to point to your frontend/SPA.
+        Method intended to be overridden in case the password reset email
+        needs to be adjusted.
         """
-        # We intentionally accept an opaque `key` on the interface here, and not
-        # implementation details such as a separate `uidb36` and `key. Ideally,
-        # this should have done on `urls` level as well.
-        path = reverse(
-            "account_reset_password_from_key", kwargs={"uidb36": "UID", "key": "KEY"}
-        )
-        path = path.replace("UID-KEY", quote(key))
-        return build_absolute_uri(self.request, path)
+        from allauth.account.internal import flows
+
+        return flows.password_reset.get_reset_password_from_key_url(self.request, key)
 
     def get_email_confirmation_url(self, request, emailconfirmation):
         """Constructs the email confirmation (activation) url.
@@ -577,22 +592,21 @@ class DefaultAccountAdapter(BaseAdapter):
         confirmations are sent outside of the request context `request`
         can be `None` here.
         """
-        url = reverse("account_confirm_email", args=[emailconfirmation.key])
-        ret = build_absolute_uri(request, url)
-        return ret
+        from allauth.account.internal import flows
 
-    def should_send_confirmation_mail(self, request, email_address, signup):
-        send_email = ratelimit.consume(
-            request,
-            action="confirm_email",
-            key=email_address.email.lower(),
+        return flows.email_verification.get_email_verification_url(
+            request, emailconfirmation
         )
-        return send_email
+
+    def should_send_confirmation_mail(self, request, email_address, signup) -> bool:
+        return True
 
     def send_account_already_exists_mail(self, email):
-        signup_url = build_absolute_uri(context.request, reverse("account_signup"))
-        password_reset_url = build_absolute_uri(
-            context.request, reverse("account_reset_password")
+        from allauth.account.internal import flows
+
+        signup_url = flows.signup.get_signup_url(context.request)
+        password_reset_url = flows.password_reset.get_reset_password_url(
+            context.request
         )
         ctx = {
             "request": context.request,
@@ -602,12 +616,20 @@ class DefaultAccountAdapter(BaseAdapter):
         self.send_mail("account/email/account_already_exists", email, ctx)
 
     def send_confirmation_mail(self, request, emailconfirmation, signup):
-        activate_url = self.get_email_confirmation_url(request, emailconfirmation)
         ctx = {
             "user": emailconfirmation.email_address.user,
-            "activate_url": activate_url,
-            "key": emailconfirmation.key,
         }
+        if app_settings.EMAIL_VERIFICATION_BY_CODE_ENABLED:
+            ctx.update({"code": emailconfirmation.key})
+        else:
+            ctx.update(
+                {
+                    "key": emailconfirmation.key,
+                    "activate_url": self.get_email_confirmation_url(
+                        request, emailconfirmation
+                    ),
+                }
+            )
         if signup:
             email_template = "account/email/email_confirmation_signup"
         else:
@@ -615,10 +637,10 @@ class DefaultAccountAdapter(BaseAdapter):
         self.send_mail(email_template, emailconfirmation.email_address.email, ctx)
 
     def respond_user_inactive(self, request, user):
-        return HttpResponseRedirect(reverse("account_inactive"))
+        return headed_redirect_response("account_inactive")
 
     def respond_email_verification_sent(self, request, user):
-        return HttpResponseRedirect(reverse("account_email_verification_sent"))
+        return headed_redirect_response("account_email_verification_sent")
 
     def _get_login_attempts_cache_key(self, request, **credentials):
         site = get_current_site(request)
@@ -696,35 +718,50 @@ class DefaultAccountAdapter(BaseAdapter):
 
     def get_login_stages(self):
         ret = []
+        ret.append("allauth.account.stages.LoginByCodeStage")
         ret.append("allauth.account.stages.EmailVerificationStage")
         if allauth_app_settings.MFA_ENABLED:
+            from allauth.mfa import app_settings as mfa_settings
+
             ret.append("allauth.mfa.stages.AuthenticateStage")
+
+            if mfa_settings.PASSKEY_SIGNUP_ENABLED:
+                ret.append("allauth.mfa.webauthn.stages.PasskeySignupStage")
         return ret
 
     def get_reauthentication_methods(self, user):
         """The order of the methods returned matters. The first method is the
         default when using the `@reauthentication_required` decorator.
         """
-        ret = []
-        if not user.is_authenticated:
-            return ret
-        if user.has_usable_password():
-            ret.append(
-                {
-                    "description": _("Use your password"),
-                    "url": reverse("account_reauthenticate"),
-                }
-            )
-        if allauth_app_settings.MFA_ENABLED:
-            from allauth.mfa.utils import is_mfa_enabled
+        from allauth.account.internal.flows.reauthentication import (
+            get_reauthentication_flows,
+        )
 
-            if is_mfa_enabled(user):
-                ret.append(
-                    {
-                        "description": _("Use your authenticator app"),
-                        "url": reverse("mfa_reauthenticate"),
-                    }
-                )
+        flow_by_id = {f["id"]: f for f in get_reauthentication_flows(user)}
+        ret = []
+        if "reauthenticate" in flow_by_id:
+            entry = {
+                "id": "reauthenticate",
+                "description": _("Use your password"),
+                "url": reverse("account_reauthenticate"),
+            }
+            ret.append(entry)
+        if "mfa_reauthenticate" in flow_by_id:
+            types = flow_by_id["mfa_reauthenticate"]["types"]
+            if "recovery_codes" in types or "totp" in types:
+                entry = {
+                    "id": "mfa_reauthenticate",
+                    "description": _("Use authenticator app or code"),
+                    "url": reverse("mfa_reauthenticate"),
+                }
+                ret.append(entry)
+            if "webauthn" in types:
+                entry = {
+                    "id": "mfa_reauthenticate:webauthn",
+                    "description": _("Use a security key"),
+                    "url": reverse("mfa_reauthenticate_webauthn"),
+                }
+                ret.append(entry)
         return ret
 
     def send_notification_mail(self, template_prefix, user, context=None, email=None):
@@ -745,13 +782,45 @@ class DefaultAccountAdapter(BaseAdapter):
             ctx.update(context)
         self.send_mail(template_prefix, email, ctx)
 
-    def generate_login_code(self):
-        forbidden_chars = "0OI18B2Z"
+    def generate_login_code(self) -> str:
+        """
+        Generates a new login code.
+        """
+        return self._generate_code()
+
+    def generate_email_verification_code(self) -> str:
+        """
+        Generates a new email verification code.
+        """
+        return self._generate_code()
+
+    def _generate_code(self):
+        forbidden_chars = "0OI18B2ZAEU"
         allowed_chars = string.ascii_uppercase + string.digits
         for ch in forbidden_chars:
             allowed_chars = allowed_chars.replace(ch, "")
         return get_random_string(length=6, allowed_chars=allowed_chars)
 
+    def is_login_by_code_required(self, login) -> bool:
+        """
+        Returns whether or not login-by-code is required for the given
+        login.
+        """
+        from allauth.account import authentication
+
+        method = None
+        records = authentication.get_authentication_records(self.request)
+        if records:
+            method = records[-1]["method"]
+        if method == "code":
+            return False
+        value = app_settings.LOGIN_BY_CODE_REQUIRED
+        if isinstance(value, bool):
+            return value
+        if not value:
+            return False
+        return method is None or method in value
+
 
 def get_adapter(request=None):
     return import_attribute(app_settings.ADAPTER)(request)
