@@ -1,4 +1,4 @@
-class AppSettings(object):
+class AppSettings:
     def __init__(self, prefix):
         self.prefix = prefix
 
@@ -56,7 +56,7 @@ class AppSettings(object):
                     "settings": app_settings,
                 }
             )
-            assert app["provider_id"]
+            assert app["provider_id"]  # nosec
             apps.append(app)
         ret["APPS"] = apps
         return ret
